#include <torch/extension.h>
#include <ATen/ATen.h>
#include <ATen/cuda/CUDAContext.h>
#include <cuda_runtime.h>
#include <vector>
#include <string>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <cmath>

namespace py = pybind11;

// Custom CUDA kernel for fused BatchNorm and ReLU
// Assumes input is a 4D tensor in NCHW format.
__global__ void fused_bn_relu_kernel(const float* __restrict__ input,
                                       float* __restrict__ output,
                                       int N, int C, int H, int W,
                                       const float* __restrict__ weight,
                                       const float* __restrict__ bias,
                                       const float* __restrict__ running_mean,
                                       const float* __restrict__ running_var,
                                       float eps) {
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int numel = N * C * H * W;
  if (idx < numel) {
    int hw = H * W;
    // Compute channel index assuming contiguous NCHW layout
    int c = (idx / hw) % C;
    float x = input[idx];
    // Use __ldg to leverage read-only cache for parameters
    float mean = __ldg(&running_mean[c]);
    float var = __ldg(&running_var[c]);
    float gamma = __ldg(&weight[c]);
    float beta = __ldg(&bias[c]);
    float norm = (x - mean) / sqrtf(var + eps);
    float y = gamma * norm + beta;
    // Fused ReLU
    output[idx] = y > 0.0f ? y : 0.0f;
  }
}

// Fused BatchNorm and ReLU forward function
at::Tensor fused_bn_relu_forward(
    at::Tensor input,
    at::Tensor weight,
    at::Tensor bias,
    at::Tensor running_mean,
    at::Tensor running_var,
    bool training,
    double momentum,
    double eps) {
  // For training, fallback to standard operations to ensure correctness.
  if (training) {
    auto bn_out = at::batch_norm(input, weight, bias, running_mean, running_var,
                                 training, momentum, eps, true);
    return at::relu(bn_out);
  }
  
  // Ensure input is contiguous for coalesced accesses
  if (!input.is_contiguous()) {
    input = input.contiguous();
  }
  // Create output tensor
  auto output = at::empty_like(input);

  // Assuming input is 4D: N x C x H x W
  int N = input.size(0);
  int C = input.size(1);
  int H = input.size(2);
  int W = input.size(3);
  int numel = input.numel();

  const int threads = 256;
  const int blocks = (numel + threads - 1) / threads;

  fused_bn_relu_kernel<<<blocks, threads, 0, at::cuda::getCurrentCUDAStream()>>>(
      input.data_ptr<float>(),
      output.data_ptr<float>(),
      N, C, H, W,
      weight.data_ptr<float>(),
      bias.data_ptr<float>(),
      running_mean.data_ptr<float>(),
      running_var.data_ptr<float>(),
      static_cast<float>(eps));
  
  // Synchronize to ensure kernel completion
  cudaDeviceSynchronize();
  return output;
}

// Dense layer function using fused BN+ReLU
at::Tensor dense_layer_fn(
    at::Tensor x,
    at::Tensor bn_weight,
    at::Tensor bn_bias,
    at::Tensor bn_mean,
    at::Tensor bn_var,
    at::Tensor conv_weight,
    bool is_training
) {
  x = fused_bn_relu_forward(x, bn_weight, bn_bias, bn_mean, bn_var,
                              is_training, 0.1, 1e-5);
  x = at::conv2d(x, conv_weight, /*bias=*/{}, /*stride=*/{1, 1}, /*padding=*/{1, 1});
  x = at::dropout(x, /*p=*/0.0, is_training);
  return x;
}

// Transition layer function using fused BN+ReLU
at::Tensor transition_layer_fn(
    at::Tensor x,
    at::Tensor bn_weight,
    at::Tensor bn_bias,
    at::Tensor bn_mean,
    at::Tensor bn_var,
    at::Tensor conv_weight,
    bool is_training
) {
  x = fused_bn_relu_forward(x, bn_weight, bn_bias, bn_mean, bn_var,
                              is_training, 0.1, 1e-5);
  x = at::conv2d(x, conv_weight);
  x = at::avg_pool2d(x, /*kernel_size=*/{2, 2}, /*stride=*/{2, 2});
  return x;
}

// Main module function implementing DenseNet121
at::Tensor module_fn(
    at::Tensor x,
    py::object params,
    bool is_training
) {
  auto get_param = [&](const std::string &key) -> at::Tensor {
    return params.attr("__getitem__")(key.c_str()).cast<at::Tensor>();
  };

  auto features_conv_weight = get_param("features_conv_weight");
  x = at::conv2d(x, features_conv_weight, /*bias=*/{}, 2, 3);

  auto features_bn_mean = get_param("features_bn_mean");
  auto features_bn_var = get_param("features_bn_var");
  auto features_bn_weight = get_param("features_bn_weight");
  auto features_bn_bias = get_param("features_bn_bias");

  x = fused_bn_relu_forward(x, features_bn_weight, features_bn_bias,
                             features_bn_mean, features_bn_var,
                             is_training, 0.1, 1e-5);
  x = at::max_pool2d(x, 3, 2, 1);

  std::vector<int> num_layers = {6, 12, 24, 16};

  for (int i = 0; i < 4; ++i) {
    std::vector<at::Tensor> features = {x};

    for (int j = 0; j < num_layers[i]; ++j) {
      std::string prefix = "block" + std::to_string(i) + "_layer" + std::to_string(j) + "_";

      auto bn_weight = get_param(prefix + "bn_weight");
      auto bn_bias = get_param(prefix + "bn_bias");
      auto bn_mean = get_param(prefix + "bn_mean");
      auto bn_var = get_param(prefix + "bn_var");
      auto conv_weight = get_param(prefix + "conv_weight");

      at::Tensor new_feature = dense_layer_fn(x, bn_weight, bn_bias, bn_mean, bn_var, conv_weight, is_training);

      features.push_back(new_feature);
      x = at::cat(features, 1);
    }

    if (i != 3) {
      std::string prefix = "transition" + std::to_string(i) + "_";

      auto bn_weight = get_param(prefix + "bn_weight");
      auto bn_bias = get_param(prefix + "bn_bias");
      auto bn_mean = get_param(prefix + "bn_mean");
      auto bn_var = get_param(prefix + "bn_var");
      auto conv_weight = get_param(prefix + "conv_weight");

      x = transition_layer_fn(x, bn_weight, bn_bias, bn_mean, bn_var, conv_weight, is_training);
    }
  }

  auto final_bn_mean = get_param("final_bn_mean");
  auto final_bn_var = get_param("final_bn_var");
  auto final_bn_weight = get_param("final_bn_weight");
  auto final_bn_bias = get_param("final_bn_bias");

  x = fused_bn_relu_forward(x, final_bn_weight, final_bn_bias,
                             final_bn_mean, final_bn_var,
                             is_training, 0.1, 1e-5);
  x = at::adaptive_avg_pool2d(x, {1, 1}).flatten(1);

  auto classifier_weight = get_param("classifier_weight");
  auto classifier_bias = get_param("classifier_bias");
  x = at::linear(x, classifier_weight, classifier_bias);

  return x;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
  m.def("forward", &module_fn, "DenseNet121 forward");
}
